with
	Ada.Text_IO,
	Ada.Strings.Unbounded,
	GNAT.Regpat,
	Regexp_List;
function Replace(Source: in String; Rule_File_Name: in String) return String is
	use Ada.Strings.Unbounded;
	
	function Convert_Pattern(Source: in String) return String is
		Result: Unbounded_String;
	begin
		Result := To_Unbounded_String("^");
		for J in Source'Range loop
			if Source(J) = '>' or else Source(J) = '<' then
				Result := Result & "\b";
			else
				Result := Result & Source(J);
			end if;
		end loop;
		--Ada.Text_IO.Put_Line(To_String(Result));
		return To_String(Result);
	end;
	
	function Convert_Character(Source: in Character) return String is
	begin
		case Source is
		when '<' =>
			return "&lt;";
		when '>' =>
			return "&gt;";
		when '&' =>
			return "&amp;";
		when Character'Val(9) =>
			return "  ";
		when others =>
			return (1 => Source);
		end case;
	end;
	
	Patterns: Regexp_List.List.List;
	Result: Unbounded_String := To_Unbounded_String("<pre>" & Character'Val(13) & Character'Val(10));
begin
	declare
		use Ada.Text_IO;
		use Regexp_List;
		use Regexp_List.List;
		use GNAT.Regpat;
		Rule_File: Ada.Text_IO.File_Type;
		Line: String(1..1024);
		Length: Natural;
		Unregistered: Cursor := No_Element;
		New_Table: Table;
		Mode: Ada.Strings.Unbounded.Unbounded_String;
	begin
		Open(Rule_File, In_File, Rule_File_Name);
		while not End_Of_File(Rule_File) loop
			Get_Line(Rule_File, Line, Length);
			if Length > 0 then
				if Line(1) = Character'Val(9) then
					Compile(New_Table.Pattern, Convert_Pattern(Line(2..Length)), Single_Line);
					New_Table.Tag_Class := Mode;
					Append(Patterns, New_Table);
				else
					if (Length = 1) and (Line(1) = '*') then
						Mode := Null_Unbounded_String;
					else
						Mode := To_Unbounded_String(Line(1..Length));
					end if;
				end if;
			end if;
		end loop;
		Close(Rule_File);
	end;

	declare
		use Regexp_List.List;
		use GNAT.Regpat;
		Processed, Next_Processed: Natural := 0;
		I: Cursor;
		Match_Result: Match_Array(0..0);
		Mode: Ada.Strings.Unbounded.Unbounded_String;
		Last: Natural := Source'Length;
	begin
		while Source(Last) = Character'Val(10) or Source(Last) = Character'Val(13) loop
			Last := Last - 1;
		end loop;
		while Processed < Last loop
			I := First(Patterns);
			while Has_Element(I) loop
				Match(Element(I).Pattern, Source(Processed + 1 .. Source'Last ), Match_Result);
				if Match_Result(0) /= No_Match then
					Next_Processed := Processed + Match_Result(0).Last - Match_Result(0).First + 1;
					
					if Element(I).Tag_Class /= Mode then
						if Mode /= Null_Unbounded_String then
							Result := Result & "</span>";
						end if;
						Mode := Element(I).Tag_Class;
						if Mode /= Null_Unbounded_String then
							Result := Result & "<span class=""" & Element(I).Tag_Class & """>";
						end if;
					end if;
					
					for K in Processed + 1 .. Next_Processed loop
						Result := Result & Convert_Character(Source(K));
					end loop;

					--Ada.Text_IO.Put_Line('[' & Source(Processed + 1 .. Next_Processed) & ']');
					Processed := Next_Processed;
					goto Next;
				end if;
				I := Regexp_List.List.Next(I);
			end loop;
			Result := Result & Convert_Character(Source(Processed + 1));
			Processed := Processed + 1;
			<<Next>>
			null;
		end loop;
		if Mode /= Null_Unbounded_String then 
			Result := Result & "</span>";
		end if;
		Result := Result & (Character'Val(13) & Character'Val(10) & "</pre>");
	end;

	return To_String(Result);
end Replace;
